#****
# [manpage_begin config n 0.5]
# [moddesc {Configuration handling for Mailroom}]
# [copyright {Copyright 2009 Jeremy Cowgar. All Rights Reserved.}]
# [category {Mailroom Configuration}]
# [keywords mailroom configuration]
# [description]
# Provides methods to deal with the configuration of Mailroom.
#

namespace eval ::mailroom::config {
  #****
  # [section Commands]
  #
  # [list_begin definitions]
  # 
  # [call [cmd {::mailroom::config load}]]
  #
  # Move all of the configuration values out of the database and into the global ::config array
  #
  proc load {} {
    global config
    variable _textwidget
    
    array set config \
      [::mailroom::db eval {SELECT name2, value FROM var WHERE name1='config'}]
    
    foreach v [array names _textwidget] {
      $_textwidget($v) delete 1.0 end
      if {[info exists config($v)]} {
        $_textwidget($v) insert end $config($v)
      }
    }
  }
  
  #****
  # [call [cmd {::mailroom::config option}] [arg group] [arg key] \
  # [opt "[option -value] [arg any]"] \
  # ]
  #
  # Retrieve or store a configuration option. Configuration options are identified by it's [arg group] and
  # [arg key] values.
  #
  # If the optional option [option -value] is supplied, then the value is stored to the database. The new
  # value is returned in this case.
  #
  proc option {name1 name2 args} {
    foreach {arg val} $args {
      switch -- $arg {
        -value { set value $val }
        default {
          error "Unknown option `$arg` sent to config_option"
        }
      }
    }
    
    if {[info exists value]} {
      ::mailroom::db eval {REPLACE INTO var (name1,name2,value) VALUES ($name1,$name2,$value)}
    } else {
      set value [::mailroom::db onecolumn {SELECT value FROM var WHERE name1=$name1 AND name2=$name2}]
    }
    
    return $value
  }
  
  #****
  # [list_end]
  #
  # [section Examples]
  #
  # [section Todo]
  #
  # [section {Bugs, Ideas, Feedback}]
  # Provide feedback to Jeremy Cowgar <jeremy@cowgar.com>
  
  namespace export option load
  namespace ensemble create
}

#****
# [manpage_end]
