#
# GUI configuration handling
#

namespace eval ::mailroom::config {
  variable _textwidget
  variable done 0
  
  # Popup the main configuration screen.
  #
  proc popup {{wait 0}} {
    variable _textwidget
    
    set w .config
    if {[winfo exists $w]} {
      wm deiconify $w
      raise $w
      load
      return
    }
    
    toplevel $w
    wm title $w {Email Configuration}
    wm iconname $w [wm title $w]
    ttk::frame $w.b
    pack $w.b -side bottom -expand 1 -fill x
    
    ttk::button $w.b.ok  -text Save   -command {
      ::mailroom::config::_save
      destroy .config
      set ::mailroom::config::done 1
    }
    ttk::button $w.b.can -text Cancel -command {
      destroy .config
      set ::mailroom::config::done 1
    }
    pack $w.b.ok $w.b.can -side left -expand 1 -pady 5
    
    ttk::frame $w.f
    pack $w.f -side top -fill both -expand 1
    set i 0

    set firstWidget 0
    foreach {name var width height} {
      {POP3 Server Hostname}     pop3host 30 1
      {POP3 Server TCP Port}     pop3port 6 1
      {POP3 Use TLS}             pop3tls 0 0
      {POP3 User Id}             pop3userid 30 1
      {POP3 Password}            pop3passwd 30 1
      {POP3 Delete}              pop3delete 0 0
      {SMTP Server Hostname}     smtphost 30 1
      {SMTP Server TCP Port}     smtpport 6 1
      {SMTP Use TLS}             smtptls 0 0
      {SMTP User Id}             smtpuserid 30 1
      {SMTP Password}            smtppasswd 30 1
      {Human-Friendly User Name} friendlyname 30 1        
      {Email Address}            fromaddr 30 1
      {Signature}                signature 50 4
      {Buckets}                  buckets 50 3
      {Debug}                    debug 0 0
    } {
      incr i
      set m $w.f.m$i
      ttk::label $m -text $name: 
      grid $m -row $i -column 1 -sticky e
      set m $w.f.e$i
      if {$i == 1} {
        set firstWidget $m
      }
      switch $height {
        0 {
          set ::config($var) 0
          ttk::checkbutton $m -variable ::config($var) -text {Yes/No}
        }
        1 {
          set ::config($var) {}
          ttk::entry $m -textvariable ::config($var) -width $width
        }
        default {
          text $m -width $width -height $height
          set _textwidget($var) $m
        }
      }
      grid $m -row $i -column 2 -sticky w -padx 3 -pady 1
    }
    load
    
    focus -force $firstWidget
    
    if {$wait} {
      vwait ::mailroom::config::done
    }
  }
  
  # Move the configuration parameters in global ::config back into
  # the database
  #
  proc _save {} {
    global config
    variable _textwidget
    
    foreach v [array names _textwidget] {
      set config($v) [$_textwidget($v) get 1.0 "end-1 char"]
    }
    ::mailroom::db transaction {
      foreach name2 [array names config] {
        set val $config($name2)
        if {$name2=="debug"} {
          set ::debug $val
        }
        ::mailroom::db eval {REPLACE INTO var(name1,name2,value) VALUES('config',$name2,$val)}
      }
    }
    
    # Update any messages that are in an invalid bucket to be in the "Inbox" state.
    set bucketNames {}
    set buckets [::mailroom::db onecolumn {SELECT value FROM var WHERE name1='config' AND name2='buckets'}]
    foreach bucket [split $buckets] {
      if {[string index $bucket 0]=="+"} {
        set bucket [string range $bucket 1 end]
      }
      lappend bucketNames $bucket
    }
    
    set bucketNamesList ('[join $bucketNames ',']')
    
    set invalidBucketCount [::mailroom::db onecolumn \
      "SELECT COUNT(msgid) FROM msg WHERE bucket NOT IN $bucketNamesList"]
    if {$invalidBucketCount > 0} {
      tk_messageBox -title "Invalid Buckets" -message "Invalid Buckets" -icon warning \
        -type ok -detail "Some messages exist in an invalid bucket. These will now\
        be placed into the 'inbox' bucket."
      ::mailroom::db eval "UPDATE msg SET bucket='inbox' WHERE bucket NOT IN $bucketNamesList" break
      message_rebuild_catalog_when_idle
    }
    
    # Update the title in case the email address or friendly name has changed.
    set_main_title
  }
  
  namespace export popup
  namespace ensemble create
}
