############################# message.tcl ##################################
# This module contains code for displaying email messages.
#

set urlreg {(http|https|ftp)\://[a-zA-Z0-9\-\.]+\.[a-zA-Z]{2,3}(:[a-zA-Z0-9]*)?/?([a-zA-Z0-9\-\._\?\,\'/\\\+&amp;%\$#\=~])*[^\.\,\)\(\s]}

# Display the message whose msgid is in the $::selected variable.
#
proc message_display {{newSelection -1}} {
  global selected
  
  if {$newSelection != -1} {
    set selected $newSelection
  }
  
  set msgid $selected
  ::mailroom::message get $msgid msg
  
  .nb.p0.t delete 1.0 end
  
  set m [mime::initialize -string $msg(b_data)] 
  if {!$msg(is_read)} {
    ::mailroom::db eval {UPDATE msg SET is_read=1 WHERE msgid=$msgid}
    .nb.p1.t item $msgid -tags {}
  }
  
  set to {}
  set cc {}
  if {![catch {mime::getheader $m To} addrs]} {
    set to [join $addrs {, }]
  }
  if {![catch {mime::getheader $m Cc} addrs]} {
    set cc [join $addrs {, }]
  }
  
  .nb.p0.t insert end "From: $msg(from_full_address)\n" header
  if {$to != {}} {
    .nb.p0.t insert end "To: $to\n" header
  }
  if {$cc != {}} {
    .nb.p0.t insert end "Cc: $cc\n" header
  }
  .nb.p0.t insert end "Sent: $msg(sent)  Rcvd: $msg(received)  Id: $msgid\n" header
  .nb.p0.t insert end "Subject: $msg(subject)\n" header
  if {$msg(attachment_count) > 0} {
    .nb.p0.t insert end "Attachments: $msg(attachment_count)  Total size: $msg(size)\n" header
  }
  .nb.p0.t insert end \n
  .nb.p0.t config -wrap word
  set body [::mailroom::message::_body $m]
  mime::finalize $m
  set body [string map [list \r { } \t {  }] $body]
  regsub -all "\[^!-~ \n\]" $body {_} body2
  
  foreach line [split $body \n] {
    set tag {}
    
    if {[string index $line 0]==">"} {
      if {[regexp {^>[ \t]?>[ \t]?>[ \t]?>[ \t]?>} $line]} {
        set tag q5
      } elseif {[regexp {^>[ \t]?>[ \t]?>[ \t]?>} $line]} {
        set tag q4
      } elseif {[regexp {^>[ \t]?>[ \t]?>} $line]} {
        set tag q3
      } elseif {[regexp {^>[ \t]?>} $line]} {
        set tag q2
      } else {
        set tag q1
      }
    }
    
    set curIdx [.nb.p0.t index insert]
    set curLine [lindex [split $curIdx .] 0]
    .nb.p0.t insert end "$line\n" $tag
    
    set matches [regexp -all -inline -indices $::urlreg $line]
    foreach {idxes a b c} $matches {
      set idxS [lindex $idxes 0]
      set idxE [lindex $idxes 1]
      .nb.p0.t tag add url $curLine.$idxS "$curLine.$idxE+1 char"
    }    
  }
  
  show_msg
  .bb.msg config -state disabled
}

# Display the html version of the message whose msgid is in
# the $::selected variable.
#

proc message_display_html {} {
  global selected
  set msgid $selected
  .nb.p2.t reset
  
  ::mailroom::message get $msgid msg

  set m [mime::initialize -string $msg(b_data)]
  .nb.p0.t config -wrap word
  set body [::mailroom::message::_body $m 1]
  mime::finalize $m
  
  if {$body==""} {
    # No HTML content must have been found. Display the text body instead.
    message_display
    .bb.html config -state disabled
  } else {
    .nb.p2.t parse $body
    show_html
    .bb.html config -state disabled
  }
}


# Display the raw text of the message whose msgid is in
# the $::selected variable.
#
proc message_display_raw {} {
  global selected
  set msgid $selected
  .nb.p0.t delete 1.0 end
  ::mailroom::message get $msgid msg
  .nb.p0.t config -wrap char
  .nb.p0.t insert end $msg(b_data)
  show_msg
  .bb.raw config -state disabled
}

# Display a MIME composition of the message whose msgid is in 
# the $::selected variable.  Hyperlinks in this display allow
# the user to view or save to disk any subpart of the MIME.
#
proc message_display_mime {} {
  global selected
  set msgid $selected
  .nb.p0.t delete 1.0 end
  ::mailroom::message get $msgid msg

  .nb.p0.t insert end "From: $msg(from_full_address)\n"
  .nb.p0.t insert end "Rcvd: $msg(received)  Id: $msgid\n"
  .nb.p0.t insert end "Subject: $msg(subject)\n\n"
  set m [mime::initialize -string $msg(b_data)]
  .nb.p0.t config -wrap char
  ::mailroom::db eval {DELETE FROM mime}
  set n 0
  _show_mime_info {} $m
  mime::finalize $m
  show_msg
  .bb.mime config -state disabled
}

proc _show_mime_info {prefix m} {
  upvar 1 n n
  if {$prefix!=""} {append prefix .}
  array set px [mime::getproperty $m]
  if {[info exists px(params)]} {
    array set mx $px(params)
    unset px(params)
  }
  foreach property [array names px] {
    if {$property=="parts"} continue
    if {$px($property)==""} continue
    .nb.p0.t insert end ${prefix}$property=[list $px($property)]\n
  }
  foreach x [array names mx] {
    if {$mx($x)==""} continue
    .nb.p0.t insert end ${prefix}$x=[list $mx($x)]\n
  }
  if {![info exists px(parts)]} {
    scan [.nb.p0.t index {end -1 chars}] %d lineno
    if {[info exists px(name)]} {
      set filename $px(name)
    } else {
      set filename {}
    }
    ::mailroom::db eval {INSERT INTO mime(lineno,partid,filename,content) VALUES($lineno,$n,$filename,$px(content))}
    incr n
    .nb.p0.t insert end ${prefix}content {} { save} mimesave " view\n" mimeview
  } else {
    set i 1
    foreach p $px(parts) {
      _show_mime_info ${prefix}m$i $p
      incr i
    }
  }
}

# This routine is called when the user clicks on the "save" or "view"
# links of a mime-view of a message.  Extract the content from the
# mime and either save it or view it.  The $destination argument 
# determines whether the result is saved to disk or displayed.  The
# $x and $y arguments are where the user clicked and are used to figure
# out which "save" or "view" hyperlink was clicked on.
#
proc mime_extract {destination x y} {
  set loc [.nb.p0.t index @$x,$y]
  set line [expr {int($loc)}]
  ::mailroom::db eval {SELECT partid AS n, content AS contentType,filename FROM mime WHERE lineno=$line} break
  global selected
  set msgid $selected
  ::mailroom::db eval {SELECT data FROM msg JOIN blob ON blobid=msgid WHERE msgid=$msgid} {
    set m [mime::initialize -string $data]
    set part [_find_subpart $m]
    if {![catch {array set px [mime::getproperty $part params]}]
         && [info exists px(name)]} {
      regsub -all {[^-a-zA-Z0-9_.]} $px(name) _ filename
    } else {
      set filename {}
    }
    set body [mime::getbody $part]
    mime::finalize $m
    break
  }
  if {![info exists body]} return
  if {$destination=="view"} {
    switch -- $contentType {
      "text/html" {
        .nb.p2.t reset
        .nb.p2.t parse -final $body
        show_html
      }
      "text/plain" {
        .nb.p0.t delete 1.0 end
        .nb.p0.t insert end $body
        show_msg
      }
      default {
        if {$filename==""} {
          tk_messageBox -icon error -title "Unknown Filename" -message "Unknown Filename" \
            -detail "Without the file name, we cannot choose the correct program to view \
              the attachment."
          return
        }
        
        
        set fullFilename [file join $::mailroom::mailroom_dir $filename]
        set out [open $fullFilename w]
        fconfigure $out -translation binary
        puts -nonewline $out $body
        close $out
        
        view_file $fullFilename
      }
    }
  } elseif {$destination=="save"} {
    if {$filename!=""} {
      set suffix [file extension $filename]
      set typename "[string toupper [string range $suffix 1 end]] Files"
      set types [list [list $typename $suffix] [list {All Files} *]]
    } elseif {$contentType=="text/html"} {
      set suffix ".html"
      set typename {HTML Files}
      set types [list [list $typename $suffix] [list {All Files} *]]
    } else {
      set types {{{All Files} *}}
    }
    set fn [tk_getSaveFile -filetypes $types -initialfile $filename]
    if {$fn!=""} {
      if {![catch {
        set out [open $fn w]
        fconfigure $out -translation binary
        puts -nonewline $out $body
      } msg]} {
        catch {close $out}
      } else {
        tk_messageBox -icon error -type ok -message $msg
      }
    }
  }
}

# Locate the n-th subpart of the MIME object $m.  $n is a variable
# in the calling context.  This routine decrements the value $n for
# the number of parts it sees.  When $n reaches zero, it returns
# that subpart.  If $n never reaches zero, this routine returns an
# empty string.
#
proc _find_subpart {m} {
  upvar 1 n n
  array set px [mime::getproperty $m]
  if {![info exists px(parts)]} {
    if {$n>0} {
      incr n -1
      return {}
    }
    return $m
  } else {
    set i 1
    foreach p $px(parts) {
      set x [_find_subpart $p]
      if {$x!=""} {return $x}
      incr i
    }
  }
  return {}
}

# Rebuild the catalog display.
#
# The $::show() variable determines which priorities are shown.
#
# The priorities are:  6=unsent 5=sent 4=unread, 3=inbox, 2=archived,
# 1=trash, 0=spam.
#
# The $::order by variable determines how the messages are sorted.
# The $::keywords variable determines any search keywords used to
# limit the display.
#
proc message_rebuild_catalog {{all 0}} {
  global show orderby keywords selected rebuild_timer limit
  catch {after cancel $rebuild_timer; unset rebuild_timer}
  .nb.p1.t delete [.nb.p1.t children {}]
  set keywords [string trim $keywords]
  if {$keywords==""} {
    set sql {
      SELECT
          m.msgid, m.bucket, datetime(m.received) AS rx, m.size, m.attachment_count,
          u.fulladdr, m.subject AS subject, '' AS snippet, m.is_read
        FROM
          msg AS m,
          user AS u
        WHERE
          m.from_userid=u.userid
    }
  } else {
    set sql {
      SELECT
          m.msgid, m.bucket, datetime(m.received) AS rx, m.size, m.attachment_count,
          u.fulladdr, m.subject AS subject,
          snippet(email) AS snippet, m.is_read
        FROM
          email AS e
          JOIN msg AS m ON (m.msgid=e.rowid)
          JOIN user AS u ON (m.from_userid=u.userid)
        WHERE
          email MATCH $keywords
    }
  }
  if {$all==0} {
    set toshow {}
    foreach i [array names show] {
      if {$show($i)} {lappend toshow $i}
    }
    if {[llength $toshow]==0} {
      return
    }
    if {[llength $toshow]==1} {
      append sql " AND m.bucket='[lindex $toshow 0]' "
    } else {
      append sql " AND m.bucket in ('[join $toshow ',']') "
    }
    if {$limit!="" && $limit!="None"} {
      append sql " AND m.received>=(SELECT julianday('now','-$limit')) "
    }
    if {$::showRead==0} {
      append sql " AND m.is_read=0"
    }
    if {$::showMine==0} {
      set myaddr [::mailroom::config option config fromaddr]
      append sql " AND u.address<>\$myaddr"
    }
  }
  switch $::orderby {
    subject {
      append sql " ORDER BY m.subject $::orderdir, m.received $::orderdir"
    }
    from {
      append sql " ORDER BY u.fulladdr $::orderdir, m.received $::orderdir"
    }
    size {
      append sql " ORDER BY m.size $::orderdir, m.received $::orderdir"
    }
    default {
      append sql " ORDER BY m.received $::orderdir"
    }
  }
  
  set cnt 0
  set first -1
  ::mailroom::db eval $sql {
    incr cnt
    if {$first == -1} {set first $msgid}
    if {!$is_read} {set tag unread} {set tag read}
    .nb.p1.t insert {} end -id $msgid -values [list $rx $subject $fulladdr [bytes_to_kb $size] $snippet] -tags $tag
  }
  
  if {![info exists selected] || $selected==""} {
    set selected $first
  }
  message_make_current $selected

  set ::status "[::number_format $cnt] matching messages"
}

proc message_rebuild_catalog_when_idle {args} {
  global rebuild_timer
  if {![info exists rebuild_timer]} {
    set rebuild_timer [after idle message_rebuild_catalog]
  }
}

# Remove the indicated message (the message id $msgid) from the catalog.
# To avoid having to rebuild the whole catalog display, we remove the
# particular row and adjust the MAP table in the ::mailroom::db database
# accordingly.
#
proc message_remove {msgid} {
  global selected
  if {$msgid!=""} {
    if {$msgid==$selected} {
      message_next
    }
    .nb.p1.t delete $msgid
  }
}

# Make the indicated message the current message in the catalog.
# Make sure it is visible in the catalog and highlight it.
#*
proc message_make_current {msgid} {
  global selected
  set selected $msgid
}

# Advance to the next message down in the catalog.
#
proc message_next {} {
  global selected
  set msgid [.nb.p1.t next $selected]
  if {$msgid==""} {
    return 0
  } else {
    message_make_current $msgid
    return 1
  }
}

# Move the selection back up to the previous message
#
proc message_previous {} {
  global selected
  set msgid [.nb.p1.t prev $selected]
  if {$msgid==""} {
    return 0
  } else {
    message_make_current $msgid
    return 1
  }
}

# Called when the mouse moves over top of the catalog.
#
proc message_mouseover {x y X Y} {
  message_cancel_popup
  if {$x<0} return
  set msgid [.nb.p1.t identify row $x $y]
  if {$msgid!=""} {
    set snippet [lindex [.nb.p1.t item $msgid -values] end]
    message_schedule_popup $msgid $snippet $X $Y
  }
}

# Cancel any existing mouseover popup and any scheduled new popup.
#
proc message_cancel_popup {} {
  global popup_timer
  catch {destroy .nb.p1.t.popup}
  catch {after cancel $popup_timer}
}

# Schedule a popup to appear if there are no mouse inputs for a
# short interval of time.
#
proc message_schedule_popup {msgid snippet X Y} {
  if {$snippet==""} return
  regsub -all {\n[ >|]*} \n$snippet { } snippet
  regsub -all {\s+} [string trim $snippet] { } snippet
  if {[string trim $snippet]==""} return
  global popup_timer
  set popup_timer [after 100 \
         [list message_show_popup $msgid $snippet $X $Y]]
}

# Draw a mouseover popup
#
proc message_show_popupxxx {msgid snippet X Y} {
  if {[winfo exists .nb.p1.t.popup]} {
    destroy .nb.p1.t.popup.t
  } else {
    toplevel .nb.p1.t.popup -bd 0 -bg black
    wm withdraw .nb.p1.t.popup
    wm overrideredirect .nb.p1.t.popup 1
  }
  message .nb.p1.t.popup.t -width 500 -font fixed -bg yellow -fg black \
     -text $snippet
  pack .nb.p1.t.popup.t -padx 1 -pady 1
  incr X 10
  incr Y 10
  wm geometry .nb.p1.t.popup +$X+$Y
  wm deiconify .nb.p1.t.popup
  raise .nb.p1.t.popup
}

proc message_show_popup {msgid snippet X Y} {
  if {[winfo exists .nb.p1.t.popup]} {
    destroy .nb.p1.t.popup.t
  } else {
    toplevel .nb.p1.t.popup -bd 0 -bg black
    wm withdraw .nb.p1.t.popup
    wm overrideredirect .nb.p1.t.popup 1
  }
  html .nb.p1.t.popup.t ;# -width 400 -height 500 -background yellow -fg black
  pack .nb.p1.t.popup.t -padx 1 -pady 1
  # puts snippet=$snippet
  .nb.p1.t.popup.t style {
     body {
       background-color: yellow;
       margin: 0;
     }
  }
  .nb.p1.t.popup.t parse -final $snippet
  .nb.p1.t.popup.t configure -width 400 -shrink 1
  update idletasks
  incr X 10
  incr Y 10
  wm geometry .nb.p1.t.popup +$X+$Y
  wm deiconify .nb.p1.t.popup
  raise .nb.p1.t.popup
}

